/*
This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef xgnu_extensions_h__
#define xgnu_extensions_h__

#if __GNUC__ > 3
#  define XGNU_ATTRIBUTE_DEFINED        1
#  define XGNU_ATTR(attr)               __attribute__(( attr ))
#  define XGNU_ATTRP(attr, ...)         __attribute__(( attr (__VA_ARGS__) ))
#  define XGNU_BI_EXPECT(expr,v)        __builtin_expect( (expr), v )
#  define XGNU_BI(bi, ...)              __builtin_ ## bi (##__VA_ARGS__)
#else /* !__GNUC__ || __GNUC__ < 4 */
#  define XGNU_ATTRIBUTE_DEFINED        0
#  define XGNU_ATTR(attr)
#  define XGNU_ATTRP(attr, ...)
#  define XGNU_BI_EXPECT(expr,v)        (expr)
#  define XGNU_BI(bi)                   _Pragma("GCC builtins not available")
#endif

#define XGNU_ALIAS(alias_name)          XGNU_ATTRP(alias, #alias_name)
#define XGNU_ALLOC_ALIGN(align)         XGNU_ATTRP(alloc_align, align)
#define XGNU_ALLOC_SIZE(x, ...)         XGNU_ATTRP(alloc_size,x, ##__VA_ARGS__)
#define XGNU_ALWAYS_INLINE              XGNU_ATTR(always_inline)
#define XGNU_FLATTEN                    XGNU_ATTR(flatten)
#define XGNU_ERROR(message)             XGNU_ATTRP(error,message)
#define XGNU_WARNING(message)           XGNU_ATTRP(warning,message)
#define XGNU_CDECL                      XGNU_ATTR(cdecl)
#define XGNU_CONST_FUNC                 XGNU_ATTR(const)
#define XGNU_PURE_FUNC                  XGNU_ATTR(pure)
#define XGNU_CONSTRUCTOR_HIGH           XGNU_ATTRP(constructor,200)
#define XGNU_DESTRUCTOR_HIGH            XGNU_ATTRP(destructor,200)
#define XGNU_CONSTRUCTOR                XGNU_ATTRP(constructor,2000)
#define XGNU_DESTRUCTOR                 XGNU_ATTRP(destructor,2000)
#define XGNU_CONSTRUCTOR_LOW            XGNU_ATTRP(constructor,20000)
#define XGNU_DESTRUCTOR_LOW             XGNU_ATTRP(destructor,20000)
#define XGNU_DEPRECATED(message)        XGNU_ATTRP(deprecated,message)
#define XGNU_DLLEXPORT                  XGNU_ATTR(dllexport)
#define XGNU_DLLIMPORT                  XGNU_ATTR(dllimport)
#define XGNU_RESOLVER_FUNC(func)        XGNU_ATTRP(ifunc,#func)
#define XGNU_MALLOC                     XGNU_ATTR(malloc)
#define XGNU_NOINLINE                   XGNU_ATTR(noinline)
#define XGNU_NONNULL(x,...)             XGNU_ATTRP(nonnull,x, ##__VA_ARGS__)
#define XGNU_RET_NONNULL                XGNU_ATTR(returns_nonnull)
#define XGNU_NORETURN                   XGNU_ATTR(noreturn)
#define XGNU_HOT                        XGNU_ATTR(hot)
#define XGNU_COLD                       XGNU_ATTR(cold)
#define XGNU_SENTINEL                   XGNU_ATTR(sentinel)
#define XGNU_UNUSED                     XGNU_ATTR(unused)
#define XGNU_USED                       XGNU_ATTR(used)

#define XGNU_ALLIGNED(align)            XGNU_ATTRP(aligned,align)

#define XGNU_VIS_DEFAULT                XGNU_ATTRP(visibility,"default")
#define XGNU_VIS_HIDDEN                 XGNU_ATTRP(visibility,"hidden")
#define XGNU_VIS_INTERNAL               XGNU_ATTRP(visibility,"internal")
#define XGNU_VIS_PROTECTED              XGNU_ATTRP(visibility,"protected")
#define XGNU_WARN_UNUSED_RES            XGNU_ATTR(warn_unused_result)
#define XGNU_CLEANUP_FUNC(func)         XGNU_ATTRP(cleanup,func)
#define XGNU_VECTOR_SIZE(size)          XGNU_ATTRP(vector_size,size)
#define XGNU_TRANSPARENT_UNION          XGNU_ATTR(transparent_union)

#define XGNU_PRINTF( format_idx, arg_idx )    \
            XGNU_ATTRP(format, printf, format_idx, arg_idx)
#define XGNU_SCANF( format_idx, arg_idx )     \
            XGNU_ATTRP(format, scanf, format_idx, arg_idx)
#define XGNU_FORMAT( arg_idx )                \
            XGNU_ATTRP(format_arg, arg_idx)

#define XGNU_LIKELY(expr)               XGNU_BI_EXPECT(expr,1)
#define XGNU_UNLIKELY(expr)             XGNU_BI_EXPECT(expr,0)
#define XGNU_UNREACHABLE                XGNU_BI(unreachable)
#define XGNU_PREFETCH(data,rw,local)    XGNU_BI(prefetch,data,rw,local)

#define XGNU_DO_PRAGMA(x)              _Pragma (#x)
#define XGNU_TODO(x)                    XGNU_DO_PRAGMA(message ("TODO - " #x))
#define XGNU_FIX(x)                     XGNU_DO_PRAGMA(message ("FIX - " #x))
#define XGNU_NO_LOOP_DEP                XGNU_DO_PRAGMA(GCC ivdep)
#endif /* xgnu_extensions_h__ */
