/*
 * This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ********
 * Allocation and die if fail functions
*/
#include "xmalloc.h"
#include "xlog.h"
#include "xgnu_extensions.h"
#include "xbacktrace.h"

#include <stddef.h>
#include <stdlib.h>

void *xmalloc(size_t size) {
  void *mem;

#ifdef XMALLOC_DEBUG
  xlog_debug_if(XGNU_UNLIKELY(size==0),"zero size alloc");
  xbacktrace();
#endif

  xlog_error_if(XGNU_UNLIKELY((mem=malloc(size))==NULL),"unable to allocate memory - size = %zd", size);

#ifdef XMALLOC_DEBUG
  xlog_debug("malloc mem at %p size %zd", mem, size);
#endif

  return mem;
}

void *xrealloc(void *mem, size_t size) {
#ifdef XMALLOC_DEBUG
  xlog_debug_if(XGNU_UNLIKELY(size==0),"realloc to size 0 -> free");
  xlog_debug_if(XGNU_UNLIKELY(mem==NULL),"realloc NULL pointer -> malloc");
  xbacktrace();
#endif

  void *ptr=realloc(mem,size);
  xlog_error_if(XGNU_UNLIKELY(ptr==NULL),"unable to realloc memory %p to size %zd", mem, size);

#ifdef XMALLOC_DEBUG
  xlog_debug_if(XGNU_UNLIKELY(ptr!=NULL),"realloc'd mem %p to %p (%zd)", mem, ptr, size);
#endif

  return ptr;
}

void xfree(void *ptr) {
#ifdef XMALLOC_DEBUG
  xlog_debug_if(XGNU_UNLIKELY(ptr==NULL),"freeing NULL ptr");
  xbacktrace();
  xlog_debug("freeing %p", ptr);
#endif

  free(ptr);
}

void xfree_and_null(void **ptr) {
#ifdef XMALLOC_DEBUG
  xlog_debug_if(XGNU_UNLIKELY(ptr==NULL),"freeing NULL *ptr");
  xbacktrace();
#endif
  if (XGNU_LIKELY(ptr!=NULL)) {
    xfree(*ptr);
    *ptr=NULL;
  }
}